// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include the GUI:
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

// Include functions to display Mat4f and Vec4f in the GUI:
# include <imgui_cg1_helpers.h>

// Include OpenGL3.3 Core functions:
#include <glad/glad.h>

// Include GLFW/OpenGL:
#include <GLFW/glfw3.h>

// Include IO utilities:
#include <iostream>

// Time utils:
#include <chrono>

// Include Vec4f class:
#include "Vec4f.h"

// Include Mat4f class:
#include "Mat4f.h"

// Include our shader wrapper class:
#include "Shader.h"

// Include our mesh class:
#include "Mesh.h"

// Include our image texture class:
#include "Texture2D.h"

// Stores the time when the program was started:
auto programStartTime = std::chrono::high_resolution_clock::now();

/**
 * This function generates and returns a tetrahedron mesh.
 */

Mesh generateTetrahedron(){
    // Create triangles for a tetrahedron:
    std::vector<Triangle> tetrahedronTriangles;
    {
        // Approximately half the size which is used to center
        // the (uncentered) coordinates in the following:
        Vec4f halfSize(sqrt(2.0f/4.0f)/2.0f,sqrt(2.0f/4.0f)/2.0f,sqrt(2.0f/4.0f)/2.0f,0.0f);

        // Vertex data for face 0 (red):
        tetrahedronTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, 0.0f) - halfSize, Vec4f(0.0f, 0.0f, -1.0f, 0.0f), Vec4f(0.0f, 1.0f, 1.0f)),
            Vertex(Vec4f(0.5f, sqrt(3.0f/4.0f), 0.0f) - halfSize, Vec4f(0.0f, 0.0f, -1.0f, 0.0f), Vec4f(0.0f, 1.0f, 1.0f)),
            Vertex(Vec4f(1.0f, 0.0f, 0.0f) - halfSize, Vec4f(0.0f, 0.0f, -1.0f, 0.0f), Vec4f(0.0f, 1.0f, 1.0f))
        ));

        // Vertex data for face 1 (green):
        tetrahedronTriangles.push_back(Triangle(
            Vertex(Vec4f(1.0f,  0.0f,  0.0f) - halfSize, Vec4f(0.82f, 0.47f, 0.33f, 0.0f), Vec4f(120.0f, 1.0f, 1.0f)),
            Vertex(Vec4f(0.5f, sqrt(3.0f/4.0f),  0.0f) - halfSize, Vec4f(0.82f, 0.47f, 0.33f, 0.0f), Vec4f(120.0f, 1.0f, 1.0f)),
            Vertex(Vec4f(0.5f,  sqrt(1.0f/12.0f), sqrt(2.0f/3.0f)) - halfSize, Vec4f(0.82f, 0.47f, 0.33f, 0.0f), Vec4f(120.0f, 1.0f, 1.0f))
        ));

        // Vertex data for face 2 (blue):
        tetrahedronTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f,  0.0f,  0.0f) - halfSize, Vec4f(-0.82f, 0.47f, 0.33f, 0.0f), Vec4f(240.0f, 1.0f, 1.0f)),
            Vertex(Vec4f(0.5f, sqrt(1.0f/12.0f),sqrt(2.0f/3.0f)) - halfSize, Vec4f(-0.82f, 0.47f, 0.33f, 0.0f), Vec4f(240.0f, 1.0f, 1.0f)),
            Vertex(Vec4f(0.5f,  sqrt(3.0f/4.0f),  0.0f) - halfSize, Vec4f(-0.82f, 0.47f, 0.33f, 0.0f), Vec4f(240.0f, 1.0f, 1.0f))
        ));

        // Vertex data for face 3 (cyan):
        tetrahedronTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f,  0.0f,  0.0f) - halfSize, Vec4f(0.0f, -0.94f, 0.34f, 0.0f), Vec4f(180.0f, 1.0f, 1.0f)),
            Vertex(Vec4f(1.0f,  0.0f,  0.0f) - halfSize, Vec4f(0.0f, -0.94f, 0.34f, 0.0f), Vec4f(180.0f, 1.0f, 1.0f)),
            Vertex(Vec4f( 0.5f,  sqrt(1.0f/12.0f), sqrt(2.0f/3.0f)) - halfSize, Vec4f(0.0f, -0.94f, 0.34f, 0.0f), Vec4f(180.0f, 1.0f, 1.0f))
        ));
    }

    return Mesh(tetrahedronTriangles);
}

/**
 * This function generates and returns a plane mesh to render water.
 */
Mesh generateWaterPlane(){
    // Create triangles for a plane:
    std::vector<Triangle> planeTriangles;
    {

        planeTriangles.push_back(Triangle(
            Vertex(Vec4f(-1.0f,-1.0f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(0,0)),
            Vertex(Vec4f( 1.0f,-1.0f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(1,0)),
            Vertex(Vec4f( 1.0f,-0.2f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(1,1))
        ));


        planeTriangles.push_back(Triangle(
            Vertex(Vec4f( 1.0f,-0.2f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(1,1)),
            Vertex(Vec4f(-1.0f,-0.2f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(0,1)),
            Vertex(Vec4f(-1.0f,-1.0f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(0,0))
        ));
    }

    return Mesh(planeTriangles);
}

/**
 * Initializes the application including the GUI which you can
 * use to test the Vec4f and Mat4f functionaliy.
 */

int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(600, 560, "Shader Framework (Computergrafik 1)", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;     // Enable Keyboard Controls

    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);

    // Setup Dear ImGui style:
    ImGui::StyleColorsDark();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf" , 16.0f);

    // Generates a tetrahedron:
    Mesh tetrahedron = generateTetrahedron();

    // Load the shader for the tetrahedron:
    Shader tetrahedronShader(
        CMAKE_SOURCE_DIR "/shader/tetrahedron.vert",
        CMAKE_SOURCE_DIR "/shader/tetrahedron.frag"
    );

    // Gerenates a plane for water:
    Mesh plane = generateWaterPlane();

    // Load the shader for the plane:
    Shader planeShader(
        CMAKE_SOURCE_DIR "/shader/plane.vert",
        CMAKE_SOURCE_DIR "/shader/plane.frag"
    );

    Texture2D waterTexture(CMAKE_SOURCE_DIR "/textures/water.jpg");

    // Main loop which is executed every frame until the window is closed:
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Start the Dear ImGui frame
        ImGui_ImplOpenGL3_NewFrame();
        ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();

        // Calculate elapsed time since program was started:
        float runTime = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::high_resolution_clock::now() - programStartTime).count() / 1000.f;

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Setup the GL viewport
        glViewport(0, 0, display_w, display_h);
        glClearColor(0.6f, 0.725f, 0.8f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Render tetrahedron:
        {
            // Activate our shader:
            tetrahedronShader.bind();

            // Upload the value of the runTime-variable to the time-variable in the shader:
            tetrahedronShader.setUniform("time", runTime);

            // Draw our triangle:
            tetrahedron.render();
        }

        // Render plane:
        {
            // Activate our shader:
            planeShader.bind();

            // Bind our water texture to texture slot 0:
            waterTexture.bind(0);

            // Tell our shader that the sampler of 'waterImage' should use texture slot 0:
            planeShader.setUniform("waterImage", 0);

            // Upload the value of the runTime-variable to the time-variable in the shader:
            planeShader.setUniform("time", runTime);

            // Draw our triangle:
            plane.render();
        }

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}

